unit AppBarForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ShellAPI;

const
  {$EXTERNALSYM ABM_SETSTATE}
  ABM_SETSTATE = $0000000A;

  WM_APPBARNOTIFY = WM_USER + 1;

type
  TAppBarPosition = (apLeft, apTop, apRight, apBottom);

  EAppBarError = class(Exception);
  TAppBarForm = class;

  TTaskbarState = (tsAutoHide, tsAlwaysOnTop);
  TTaskbarStates = set of TTaskbarState;

  TTaskbarStateChangeEvent = procedure (Sender: TObject; State: TTaskbarStates) of object;
  TFullScreenAppEvent = procedure (Sender: TObject; Opened : Boolean) of object;

  TWindowsTaskBar = class(TPersistent)
  private
    function GetPosition: TRect;
    function GetState: TTaskBarStates;
    procedure SetState(const Value: TTaskBarStates);
  protected
    function AppBarGetTaskbarState : TTaskbarStates;
    function AppBarGetTaskBarPos(var ARect : TRect) : Boolean;
    procedure AppBarSetState(AState : TTaskBarStates);
  public
    property Position : TRect read GetPosition;
  published
    property State : TTaskBarStates read GetState write SetState;
  end;

  TAppBarForm = class(TForm)
  private
    FRegistered: Boolean;
    FResizing : Boolean;
    FBarPosition: TAppBarPosition;
    FSavedPos : TRect;
    FSavedStyle : DWORD;
    FOnTaskbarStateChange: TTaskbarStateChangeEvent;
    FBeforeArrange: TNotifyEvent;
    FAfterArrange: TNotifyEvent;
    FOnFullScreenApp: TFullScreenAppEvent;
    FTaskBar: TWindowsTaskbar;
    { Private declarations }
    procedure AppBarCallBack(var Msg : TMessage); message WM_APPBARNOTIFY;
    procedure WMActivate(var Msg : TWMActivate); message WM_ACTIVATE;
    procedure WMPositionChanged(var Msg : TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure WMExitSizeMove(var Msg : TMessage); message WM_EXITSIZEMOVE;
    procedure WMNCHitTest(var Msg : TWMNcHitTest); message WM_NCHITTEST;
    procedure SetRegistered(const Value: Boolean);
    procedure SetBarPosition(const Value: TAppBarPosition);
  protected
    { Protected declarations }
    procedure FillDefault(var AData : TAppBarData);
    procedure SaveState;
    procedure RestoreState;
    procedure DoTaskbarStateChange(State : TTaskbarStates); dynamic;
    procedure DoFullScreenApp(Opened : Boolean);
    procedure DoBeforeArrange; dynamic;
    procedure DoAfterArrange; dynamic;
    // SHAppBarMessage wrapers
    procedure AppBarActivate;
    function AppBarGetAutoHide(AtEdge : TAppBarPosition) : HWND;
    function AppBarSetAutoHide(ARegister : Boolean) : Boolean;
    function AppBarNew : Boolean;
    function AppBarRemove : Boolean;
    procedure AppBarQueryPos(AEdge : TAppBarPosition; var ARect : TRect);
    procedure AppBarSetPos(AEdge : TAppBarPosition; var ARect : TRect);
    procedure AppBarWindowPosChanged;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    { Published declarations }
    property AppBarPosition : TAppBarPosition read FBarPosition write SetBarPosition;
    property Registered : Boolean read FRegistered write SetRegistered;
    property Taskbar : TWindowsTaskbar read FTaskBar;
    property OnTaskbarStateChange : TTaskbarStateChangeEvent read FOnTaskbarStateChange write FOnTaskbarStateChange;
    property OnFullScreenApp : TFullScreenAppEvent read FOnFullScreenApp write FOnFullScreenApp;
    property AfterArrange : TNotifyEvent read FAfterArrange write FAfterArrange;
    property BeforeArrange : TNotifyEvent read FBeforeArrange write FBeforeArrange;
  end;

procedure Register;

implementation

//{$R *.dfm}

uses DsgnIntf;

resourcestring
  SRegisterFailed       = 'Registration of the appbar failed';
  SUnRegisterFailed     = 'Remove of the appbar failed';
  SAutoHideAlreadyReg   = 'There is already a registered auto-hide bar on this edge of the screen';
  SAutoHideUnRegFailed  = 'Remove of the auto-hide appbar failed';

procedure Register;
begin
  RegisterCustomModule(TAppBarForm, TCustomModule);
end;

{ TAppBarForm }

procedure TAppBarForm.AppBarCallBack(var Msg: TMessage);
var
  Data : TAppBarData;
  NativeState : UINT;
  State : TTaskbarStates;
begin
  FillDefault(Data);
  
  case Msg.WParam of
    ABN_STATECHANGE :
      begin
        NativeState := SHAppBarMessage(ABM_GETSTATE,data);

        State := [];
        if (NativeState and ABS_AUTOHIDE) = ABS_AUTOHIDE then
          Include(State,tsAutoHide);
        if (NativeState and ABS_ALWAYSONTOP) = ABS_ALWAYSONTOP then
          Include(State,tsAlwaysOnTop);

        DoTaskbarStateChange(State);
      end;
    ABN_POSCHANGED : SetBounds(Left, Top, Width, Height);
    ABN_FULLSCREENAPP :
      begin
        if Msg.LParam = 1 then SetWindowPos(Handle,HWND_BOTTOM,0,0,0,0,SWP_NOMOVE + SWP_NOSIZE + SWP_NOACTIVATE)
                          else
                            if FormStyle = fsStayOnTop then
                              SetWindowPos(Handle,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE + SWP_NOSIZE + SWP_NOACTIVATE);

        DoFullScreenApp(Msg.LParam = 1);
      end;
    ABN_WINDOWARRANGE :
      if Msg.LParam = 1 then DoBeforeArrange
                        else DoAfterArrange;
  end;
end;

function TAppBarForm.AppBarGetAutoHide(AtEdge: TAppBarPosition): HWND;
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  Data.uEdge := Cardinal(AtEdge);
  Result := SHAppBarMessage(ABM_GETAUTOHIDEBAR,Data);
end;

function TAppBarForm.AppBarSetAutoHide(ARegister : Boolean) : Boolean;
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  Data.uEdge := Cardinal(AppBarPosition);
  Data.lParam := Integer(ARegister);
  Result := SHAppBarMessage(ABM_SETAUTOHIDEBAR,Data) > 0;
end;

function TAppBarForm.AppBarNew: Boolean;
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  Data.uCallbackMessage := WM_APPBARNOTIFY;
  Result := SHAppBarMessage(ABM_NEW,Data) > 0;
end;

procedure TAppBarForm.AppBarQueryPos(AEdge : TAppBarPosition; var ARect: TRect);
var
  Data : TAppBarData;
  TempWidth, TempHeight : Integer;
begin
  FillDefault(Data);
  Data.uEdge := Cardinal(AEdge);

  TempWidth := ARect.Right - ARect.Left;
  TempHeight := ARect.Bottom - ARect.Top;
  if (FBarPosition in [apLeft, apRight]) then
    begin
      if FBarPosition = apLeft then
        begin
          ARect.Left := 0;
          ARect.Right := TempWidth;
        end
        else
        begin
          ARect.Left := Screen.Width - TempWidth;
          ARect.Right := Screen.Width;
        end;

      ARect.Top := 0;
      ARect.Bottom := Screen.Height;
    end
    else
    begin
      if FBarPosition = apTop then
        begin
          ARect.Top := 0;
          ARect.Bottom := TempHeight;
        end
        else
        begin
          ARect.Top := Screen.Height - TempHeight;
          ARect.Bottom := Screen.Height;
        end;

      ARect.Left := 0;
      ARect.Right := Screen.Width;
    end;

  Data.rc := ARect;
  SHAppBarMessage(ABM_QUERYPOS,Data);
  ARect := Data.rc;

  case FBarPosition of
    apLeft   : ARect.Right := ARect.Left + TempWidth;
    apRight  : ARect.Left := ARect.Right - TempWidth;
    apTop    : ARect.Bottom := ARect.Top + TempHeight;
    apBottom : ARect.Top := ARect.Bottom - TempHeight;
  end;
end;

function TAppBarForm.AppBarRemove : Boolean;
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  // This should always return true
  Result := SHAppBarMessage(ABM_REMOVE,Data) > 0;
end;

procedure TAppBarForm.AppBarSetPos(AEdge : TAppBarPosition; var ARect: TRect);
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  Data.uEdge := Cardinal(AEdge);
  Data.rc := ARect;
  SHAppBarMessage(ABM_SETPOS,Data);
  ARect := Data.rc;
end;

procedure TAppBarForm.AppBarWindowPosChanged;
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  SHAppBarMessage(ABM_WINDOWPOSCHANGED,Data);
end;

constructor TAppBarForm.Create(AOwner: TComponent);
begin
  FBarPosition := apTop;
  FRegistered := false;
  FResizing := false;
  FTaskBar := TWindowsTaskBar.Create;
  inherited Create(AOwner);
end;

procedure TAppBarForm.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);

  Params.ExStyle := (Params.ExStyle or WS_EX_TOOLWINDOW) and not WS_EX_APPWINDOW;
end;

destructor TAppBarForm.Destroy;
begin
  SetRegistered(false);
  FTaskBar.Free;
  inherited;
end;

procedure TAppBarForm.DoAfterArrange;
begin
  if Assigned(FAfterArrange) then
    FAfterArrange(Self);
end;

procedure TAppBarForm.DoBeforeArrange;
begin
  if Assigned(FBeforeArrange) then
    FBeforeArrange(Self);
end;

procedure TAppBarForm.DoFullScreenApp(Opened: Boolean);
begin
  if Assigned(FOnFullScreenApp) then
    FOnFullScreenApp(Self,Opened);
end;

procedure TAppBarForm.DoTaskbarStateChange(State: TTaskbarStates);
begin
  if Assigned(FOnTaskbarStateChange) then
    FOnTaskbarStateChange(Self,State);
end;

procedure TAppBarForm.RestoreState;
begin
  SetBounds(FSavedPos.Left, FSavedPos.Top, FSavedPos.Right, FSavedPos.Bottom);

  SetWindowLong(Handle,GWL_STYLE,FSavedStyle);
  SetWindowPos(Handle , 0, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE or SWP_NOZORDER or SWP_NOACTIVATE or SWP_DRAWFRAME);
end;

procedure TAppBarForm.SaveState;
begin
  FSavedPos.Left := Left;
  FSavedPos.Top := Top;
  FSavedPos.Right := Width;
  FSavedPos.Bottom := Height;

  FSavedStyle := GetWindowLong(Handle,GWL_STYLE);
end;

procedure TAppBarForm.SetBarPosition(const Value: TAppBarPosition);
var
  ReRegister : Boolean;
begin
  ReRegister := Registered;
  Registered := false;

  FBarPosition := Value;

  if ReRegister then
    Registered := true;

  SetBounds(Left, Top, Width, Height);
end;

procedure TAppBarForm.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  R : TRect;
begin
  if not FRegistered then
      inherited
    else
    begin
      R := Rect(Left, Top, Left + Width, Top + Height);

      AppBarQueryPos(FBarPosition, R);
      AppBarSetPos(FBarPosition, R);

      MoveWindow(Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top, TRUE);
    end;
end;

procedure TAppBarForm.SetRegistered(const Value: Boolean);
begin
  if csDesigning in ComponentState then
    exit;

  if Value <> FRegistered then
    begin
      if Value then
        begin
          if not AppBarNew then
            raise EAppBarError.Create(SRegisterFailed);

          SaveState;
        end
        else
        begin
          if not AppBarRemove then
            raise EAppBarError.Create(SUnRegisterFailed);
        end;

      FRegistered := Value;
      if FRegistered then
        begin
          SetWindowLong(Handle,GWL_STYLE,FSavedStyle and (not WS_CAPTION));
          SetBounds(Left, Top, Width, Height);
        end
        else
          RestoreState;
    end;
end;

procedure TAppBarForm.WMActivate(var Msg: TWMActivate);
begin
  if FRegistered then
    AppBarActivate;

  inherited;
end;

procedure TAppBarForm.WMPositionChanged(var Msg: TWMWINDOWPOSCHANGED);
begin
  if FRegistered then
    AppBarWindowPosChanged;

  inherited;
end;

procedure TAppBarForm.WMExitSizeMove(var Msg: TMessage);
begin
  if FRegistered then
    SetBounds(Left, Top, Width, Height);

  inherited;
end;

procedure TAppBarForm.WMNCHitTest(var Msg: TWMNcHitTest);
begin
  inherited;

  if FRegistered and (Msg.Result in [HTLEFT..HTBOTTOMRIGHT]) then
    begin
      case FBarPosition of
        apLeft : if Msg.Result <> HTRIGHT then
                   Msg.Result := HTCLIENT;
        apTop : if Msg.Result <> HTBOTTOM then
                  Msg.Result := HTCLIENT;
        apRight : if Msg.Result <> HTLEFT then
                    Msg.Result := HTCLIENT;
        apBottom : if Msg.Result <> HTTOP then
                     Msg.Result := HTCLIENT;
      end;
    end;
end;

procedure TAppBarForm.AppBarActivate;
var
  Data : TAppBarData;
begin
  FillDefault(Data);
  SHAppBarMessage(ABM_ACTIVATE,Data);
end;

procedure TAppBarForm.FillDefault(var AData: TAppBarData);
begin
  AData.cbSize := SizeOf(TAppBarData);
  AData.hWnd := Handle;
end;

{ TWindowsTaskBar }

function TWindowsTaskBar.AppBarGetTaskBarPos(var ARect : TRect) : Boolean;
var
  Data : TAppBarData;
begin
  Data.cbSize := SizeOf(TAppBarData);
  Data.hWnd := GetDesktopWindow;
  Result := SHAppBarMessage(ABM_GETTASKBARPOS,Data) > 0;
  ARect := Data.rc;
end;

function TWindowsTaskBar.AppBarGetTaskbarState: TTaskbarStates;
var
  Data : TAppBarData;
  NativeState : UINT;
begin
  Data.cbSize := SizeOf(TAppBarData);
  NativeState := SHAppBarMessage(ABM_GETSTATE,data);

  Result := [];
  if (NativeState and ABS_AUTOHIDE) = ABS_AUTOHIDE then
    Include(Result,tsAutoHide);
  if (NativeState and ABS_ALWAYSONTOP) = ABS_ALWAYSONTOP then
    Include(Result,tsAlwaysOnTop);
end;

procedure TWindowsTaskBar.AppBarSetState(AState: TTaskBarStates);
const
  BAR_STATES : array[TTaskBarState] of LPARAM = (ABS_AUTOHIDE, ABS_ALWAYSONTOP);
var
  Data : TAppBarData;
  NewState : LPARAM;
  i : TTaskbarState;
begin
  Data.cbSize := SizeOf(TAppBarData);
  Data.hWnd := GetDesktopWindow;
  NewState := 0;
  for i := Low(TTaskBarState) to High(TTaskBarState) do
    if i in AState then
      NewState := NewState or BAR_STATES[i];

  Data.lParam := NewState;
  SHAppBarMessage(ABM_SETSTATE,Data);
end;

function TWindowsTaskBar.GetPosition: TRect;
var
  R : TRect;
begin
  R := Rect(0,0,0,0);
  Result := R;

  if AppBarGetTaskBarPos(R) then
    Result := R;
end;

function TWindowsTaskBar.GetState: TTaskBarStates;
begin
  Result := AppBarGetTaskbarState;
end;

procedure TWindowsTaskBar.SetState(const Value: TTaskBarStates);
begin
  AppBarSetState(Value);
end;

end.
